c
c   This file contains a very simple program for testing various combinations
c   of the 4-d grid operations.  The idea is to initialize each segment with
c   the values of the global indices, with ghost points being negative.
c   After a copy, the values are tested.  Note that because the ghost-points
c   exist at the periodic physical boundaries and not at the non-periodic
c   physical boundaries, we can test both cases.
c
c     int mdim,          /* size of array in this dimension */
c     int is_parallel,   /* true if the array is parallelized in
c                           this dimension */
c     int start,         /* starting index for local piece */
c     int end,           /* ending index for local piece */
c     int loc            /* location in this dimension of the 
c                           processor */
c     int ndim           /* number of processors in this dimension,
c 	                   -1 if unspecified */
c     int sg             /* start and end ghost limits */
c     int eg             /* start and end ghost limits */
c     int nsg            /* start and end ghost limits for neighbor */
c     int neg
c
       integer function worker()
       integer iii, k, mx, my, mz, mq, nd
       parameter(nd=4, NBYTES=8)
       integer maxnx, maxny, maxnz, maxnq
       parameter(maxnx = 16, maxny = 16, maxnz=16, maxnq=16 )
c      Fields in the "SZ" array 
       include '../meshf.h'
       integer pimytid, pinumtids
c      integer pigtoken
       integer pgm, szar(0:9,0:nd-1), iper(0:nd-1)
       integer nx, ny, nz, nq
       integer sx,sxgp,ex,exgp,sy,sygp,ey,eygp,sz,szgp,ez,ezgp
       integer sq, sqgp, eq, eqgp
       double precision a(maxnx*maxny*maxnz*maxnq)
       integer myid, nproc
       integer err
       logical printar, printfinal
c       integer argc, argv
c
c Define the domain as a 2-d mesh of size mx x my, to be subdivided 
c in both dimensions
c
c       mx    = 256
c       my    = 256
c       mz    = 16
c       printar = .false.
c
        mx = 2
        my = 4
        mz = 2
        mq = 3
c        printar = .true.
        printar = .false.
        printfinal = .false. 
c       call syargload( argc, argv )
c      if (syarghasname( argc, argv, 1,"-periodic" )) print *, 'found'
c
c
c      ? a loop over boundary cases
       do 1000 iii=0, 15
c      
       szar(szmdim,0)       = mx
       szar(szisparallel,0) =  1
       szar(szndim,0)       = -1
       szar(szmdim,1)       = my
       szar(szisparallel,1) =  1
       szar(szndim,1)       = -1
       szar(szmdim,2)       = mz
       szar(szisparallel,2) =  1
       szar(szndim,2)       = -1
       szar(szmdim,3)       = mq
       szar(szisparallel,3) =  1
       szar(szndim,3)       = -1
       iper(0)            = mod( iii, 2 )
       iper(1)            = mod( iii/2, 2 )
       iper(2)            = mod( iii/4, 2 )
       iper(3)            = mod( iii/8, 2 )
c       call syarggetintvec( argc, argv, 1, "-iper", 3, iper )
       myid  = pimytid()
       nproc = pinumtids()
       if (myid .eq. 0) print *, 'Iper = ', iper
c
c      Build the communications pattern by:
c 
c      1. Compute the size of the ghost-points from the computational 
c         stencil
c      2. Compute the local part of the array that this processor 
c         is responsible for
c      3. Build the communication pattern and "compile" it
c
c      Setup the ghost points from the stencil 
c      Note that this assumes doublely periodic boundary conditions.
       call BCFindGhostFromStencil( nd, szar, 0, 0, boxstencil )

       call BCGlobalToLocalArray( nd, szar, nproc, myid )
       call BCSetGhostWidths( nd, szar, iper )
c
c       do 10 i=0, nproc
c       if (pigtoken(0,i).ne.0) then
c          print *, 'Processor ', i
c       call BCPrintArrayPartVerbose( 0, nd, szar )
c       endif
c 10    continue
       pgm = BCBuildArrayPGM( nd, szar, nproc, myid, NBYTES )
       call BCUseOrderedSend( pgm )
       call BCArrayCompile( pgm, 0 )
c       call bcprint_pgms( pgm, 0 )
c
c      Compute the parameters of our part of the domain
c
       sx   = szar(szstart,0) + 1
       ex   = szar(szend,0) + 1
       sxgp = szar(szsg,0)
       exgp = szar(szeg,0)
       nx   = ex - sx + 1 + sxgp + exgp
       sy   = szar(szstart,1) + 1
       ey   = szar(szend,1) + 1
       sygp = szar(szsg,1)
       eygp = szar(szeg,1)
       ny   = ey - sy + 1 + sygp + eygp
       sz   = szar(szstart,2) + 1
       ez   = szar(szend,2) + 1
       szgp = szar(szsg,2)
       ezgp = szar(szeg,2)
       nz   = ez - sz + 1 + szgp + ezgp
       sq   = szar(szstart,3) + 1
       eq   = szar(szend,3) + 1
       sqgp = szar(szsg,3)
       eqgp = szar(szeg,3)
       nq   = eq - sq + 1 + sqgp + eqgp
c
c      Sanity check on subproblem size
       if (ny * nx * nz * nq .gt. maxnx * maxny * maxnz * maxnq) then
	   print *, 'Selected domain too large '
	   worker = 1
	   return
       endif
c
c      initialize a 
       call InitDomain( a, mx,my,mz,mq,
     *                  sx,sxgp,ex,exgp,sy,sygp,ey,eygp,
     *                  sz,szgp,ez,ezgp,sq,sqgp,eq,eqgp)
       if (printar) then
       if (myid .eq. 0) print *, 'Initial arrays:'
           do k=1,nz
           print *, 'plane ', k
          call dispall( a(1+(k-1)*(nx*ny)), 
     *                  sx, sxgp, ex, exgp, sy, sygp, ey, eygp, 
     *                  mx, my, 16 )
          enddo
       endif
cc
       call BCexec( pgm, a, a )
c
c      Check results
       call CheckValues( a, mx,my,mz,mq,
     *                   sx,sxgp,ex,exgp,sy,sygp,ey,eygp, 
     *                   sz,szgp,ez,ezgp,sq,sqgp,eq,eqgp,iper, err )
       call PIgisum( err, 1, i, 0 )
       if (printfinal .or. (err .gt. 0 .and. printar)) then 
          do k=1,nz
           print *, 'plane ', k
          call dispall( a(1+(k-1)*(nx*ny)), 
     *                  sx, sxgp, ex, exgp, sy, sygp, ey, eygp, 
     *                   mx, my, 16 )
          enddo
       endif
c
       call BCfree( pgm )
c
 1000  continue
       worker = 0
       return
       end
c
       subroutine InitDomain( a, mx,my,mz,mq,
     +                           sx,sxgp,ex,exgp,sy,sygp,ey,eygp,
     +                           sz,szgp,ez,ezgp,sq,sqgp,eq,eqgp)
       integer      mx,my,sx,sxgp,ex,exgp,sy,sygp,ey,eygp
       integer      mq,sq,sqgp,eq,eqgp
       integer      mz,sz,szgp,ez,ezgp
       double precision a(sx-sxgp:ex+exgp,sy-sygp:ey+eygp,
     +                    sz-szgp:ez+ezgp,sq-sqgp:eq+eqgp)
       integer i, j, k, m
c
c     Set everything to - global index
c
      do m=sq-sqgp,eq+eqgp
        do k=sz-szgp,ez+ezgp
          do j = sy-sygp,ey+eygp
            do i = sx-sxgp,ex+exgp
              a(i,j,k,m) = -( i+mx * ((j-1)+my * ((k-1)+mq*(m-1))))
            enddo
          enddo
        enddo
       enddo
c
c ---  initialize the interior as the local indices
c
       do m = sq, eq
         do k = sz, ez
           do j = sy, ey
             do i = sx, ex
               a(i,j,k,m) = i + mx * ((j-1)+my * ( (k-1) + mq * (m-1)))
             enddo
          enddo
        enddo
       enddo
       return
       end
c
       subroutine CheckValues( a, mx,my,mz,mq,
     +                      sx,sxgp,ex,exgp,sy,sygp,ey,eygp,
     +                      sz,szgp,ez,ezgp,sq,sqgp,eq,eqgp,iper, err)
       integer      mx,my,sx,sxgp,ex,exgp,sy,sygp,ey,eygp,iper(0:2)
       integer      mz,sz,szgp,ez,ezgp,mq,sq,sqgp,eq,eqgp
       double precision a(sx-sxgp:ex+exgp,sy-sygp:ey+eygp,
     +                    sz-szgp:ez+ezgp,sq-sqgp:eq+eqgp)
       integer err, tval, i, j, k, m, cnt
       integer PImytid
c
c     Set everything should now be global index
c
       err = 0
       cnt = 0
       do m = sq-sqgp, eq+eqgp
         do k = sz-szgp, ez+ezgp
           do j = sy-sygp,ey+eygp
              do i = sx-sxgp,ex+exgp
c This version should be correct for periodic boundaries
             tval = 1+mod(mx+i-1,mx) + mx * ( mod(my+j-1,my) + 
     *              my * (mod(mz+k-1,mz) + mq * mod(mq+m-1,mq)))
             if (abs(a(i,j,k,m) - tval) .gt. 1.0e-10) then
                 err = err + 1
                 if (err .gt. 10) return 
                 write (*,1) PImytid(), i, j, k, m, a(i,j,k,m ), tval
 1               format( ' [', i2, '] (', i3, ',', i5 ,',', i5, 
     *                   ',', i5, ') = ', f6.0, ', should be ', i4 )
             endif
             cnt = cnt + 1
             enddo
          enddo
         enddo
       enddo
       call pigisum( cnt, 1, k, 0 )
       if (pimytid() .eq. 0) then
          print *, 'Tested ', cnt, ' values without error'
       endif
       return 
       end
